;---------------------------------------------------------------------
; DMK Virtual Disk Interface                     (C) 2009 Miguel Dutra
;---------------------------------------------------------------------

;---------------------------------------------------------------------
; DMK Disk Header Flags
;---------------------------------------------------------------------

DMK_Flag_SingleSided    equ         00010000y                           ; 1:Single Sided
DMK_Flag_SingleDensity  equ         01000000y                           ; 1:Single Density, 0:Double Density
DMK_Flag_IgnoreDensity  equ         10000000y                           ; 1:Ignore Density (disk is DD but should always read 1 byte)

;---------------------------------------------------------------------
; DMK IDAM Pointer Masks
;---------------------------------------------------------------------

DMK_IDAM_Density        equ         1000000000000000y                   ; Density flag (0:Single, 1:Double)
DMK_IDAM_Undefined      equ         0100000000000000y                   ; Undefined
DMK_IDAM_Offset         equ         0011111111111111y                   ; Pointer offset to the FEh byte of the IDAM

.data?

;---------------------------------------------------------------------
; DMK Disk Header 
;---------------------------------------------------------------------

DMK_Header  struct
    WP                  BYTE        ?                                   ; Write-Protected disk (00:No, FF:Yes)
    Tracks              BYTE        ?                                   ; Number of tracks on virtual disk
    TrackLength         WORD        ?                                   ; Track length
    Flags               BYTE        ?                                   ; Virtual disk option flags
    Reserved            BYTE        7 dup (?)                           ; Reserved for future options
    Signature           DWORD       ?                                   ; 0h:Virtual disk in native format, 12345678h:Real disk specification
DMK_Header  ends

;---------------------------------------------------------------------
; DMK Track Header 
;---------------------------------------------------------------------

DMK_TrackHeader struct
    IDAM_PTR            WORD        64 dup (?)                          ; Each track contains a header with up to 64 pointers to seactor's IDAMs
DMK_TrackHeader ends

;---------------------------------------------------------------------
; DMK Sector ID Field
;---------------------------------------------------------------------

DMK_SectorID    struct
    IDAM                BYTE        ?                                   ; ID Address Mark (FEh)
    Track               BYTE        ?
    Side                BYTE        ?
    Sector              BYTE        ?
    SectorSize          BYTE        ?
    CRC1                BYTE        ?
    CRC2                BYTE        ?
DMK_SectorID    ends

;---------------------------------------------------------------------
; DMK Sector Header
;---------------------------------------------------------------------

DMK_SectorHeader    struct
    SectorID            DMK_SectorID    <>                              ; Sector ID Field
    IDAM_PTR            WORD            ?                               ; Original pointer offset to the sector IDAM
    SectorSize          WORD            ?                               ; Sector size (128, 256, 512 or 1024)
    DoubleByte          BYTE            ?                               ; Double byte flag
    pIDAM               DWORD           ?                               ; Pointer to the sector ID Address Mark
    pDAM                DWORD           ?                               ; Pointer to the sector Data Address Mark
DMK_SectorHeader    ends

;---------------------------------------------------------------------
; DMK Class
;---------------------------------------------------------------------

CDMK    struct
                        CVDI        <>                                  ; Inherit all VDI class members
    Header              DMK_Header  <>                                  ; DMK disk header
    pTrack              DWORD       ?                                   ; Pointer to track data buffer
    LoadedTrack         WORD        ?                                   ; Cylinder and Head of the currently loaded track
    CacheWrite          BYTE        ?                                   ; Indicates the need to write the cache contents back to disk
    iSides              BYTE        ?                                   ; Number of allocated (in opposition to actually formatted) disk sides
CDMK    ends

.data

;---------------------------------------------------------------------
; DMK Virtual Table
;---------------------------------------------------------------------

DMK_VTBL    struct                                                      ; Declare the structure and initializers
    Load                DWORD       DMK_Load                            ; VDI-derived functions
    Unload              DWORD       DMK_Unload
    Read                DWORD       DMK_Read
    Write               DWORD       DMK_Write
    GetTracks           DWORD       VDI_GetTracks
    GetSides            DWORD       VDI_GetSides
    GetSPT              DWORD       VDI_GetSPT
    GetDensity          DWORD       VDI_GetDensity
    GetFirstTrack       DWORD       VDI_GetFirstTrack
    GetLastTrack        DWORD       VDI_GetLastTrack
    GetFirstSide        DWORD       VDI_GetFirstSide
    GetLastSide         DWORD       VDI_GetLastSide
    GetFirstSector      DWORD       VDI_GetFirstSector
    GetLastSector       DWORD       VDI_GetLastSector
    GetDiskGeometry     DWORD       DMK_GetDiskGeometry                 ; DMK-specific functions
    Seek                DWORD       DMK_Seek
    LoadTrack           DWORD       DMK_LoadTrack
    SaveTrack           DWORD       DMK_SaveTrack
    SeekIDAM            DWORD       DMK_SeekIDAM
    GetSectorID         DWORD       DMK_GetSectorID
    PutSectorID         DWORD       DMK_PutSectorID
    SeekDAM             DWORD       DMK_SeekDAM
    GetSectorData       DWORD       DMK_GetSectorData
    PutSectorData       DWORD       DMK_PutSectorData
    UpdateCRC           DWORD       DMK_UpdateCRC
DMK_VTBL    ends

DMK                     DMK_VTBL    <>                                  ; Define a variable of that type

.code
    
;---------------------------------------------------------------------
; DMK_Load                                        Initialize interface
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; hFile:    Handle to an open Virtual Floppy Disk file
; Flags:    User-defined behavior modification flags
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_Load    proc    pDMK:DWORD, hFile:HANDLE, Flags:DWORD

            local   Header:DMK_Header
            local   Bytes:DWORD
            local   pTrack:DWORD

    ; --- Load DMK disk header ---

            fn      SetFilePointer, hFile, 0, NULL, FILE_BEGIN          ; Set pointer at beginning of file

            cmp     eax, -1
            jne     @F

            fn      GetLastError
            jmp     Done
            
@@:         fn      ReadFile, hFile, addr Header, sizeof Header, addr Bytes, NULL

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done
            
    ; --- Validate the header ---

@@:         cmp     Header.WP, 00h                                      ; Write-Protected flag must be either 00 or FF
            je      @F

            cmp     Header.WP, 0FFh
            jne     Error

@@:         cmp     Header.Tracks, 35                                   ; Number of tracks must be at least 35
            jb      Error
            
            cmp     Header.Signature, 00000000h                         ; Signature must be all zeroes for a virtual disk
            jne     Error

;            movzx   eax, Header.Tracks                                  ; Disk size = Tracks * TrackLength * Sides + (sizeof DMK_Header)
;            test    Header.Flags, DMK_Flag_SingleSided
;            setz    cl
;            shl     eax, cl
;            movzx   edx, Header.TrackLength
;            mul     edx
;            add     eax, sizeof DMK_Header

;            push    eax
;            fn      GetFileSize, hFile, NULL                            ; Get file size
;            pop     edx

;            cmp     eax, edx                                            ; File size must match or exceed the disk size
;            jb      Error

    ; --- Allocate memory for a track ---

            movzx   eax, Header.TrackLength

            fn      GlobalAlloc, GPTR, eax

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     pTrack, eax

    ; --- Update member variables ---

            mov     edx, pDMK

            mov     eax, hFile                                          ; Update handle to file
            mov     (CDMK PTR [edx]).hFile, eax

            mov     eax, Flags                                          ; Update the behavior modification flags
            mov     (CDMK PTR [edx]).Flags, eax

            mov     eax, pTrack                                         ; Update pointer to track data buffer
            mov     (CDMK PTR [edx]).pTrack, eax

            mov     (CDMK PTR [edx]).LoadedTrack, 0FFFFh                ; Initialize the track caching control
            mov     (CDMK PTR [edx]).CacheWrite, 0

            fn      MemCopy, addr Header, addr (CDMK PTR [edx]).Header, sizeof Header    ; Update the disk header

    ; --- Find out the disk geometry ---

            mcall   DMK_VTBL, GetDiskGeometry, pDMK
            jmp     Done

Error:      mov     eax, ERROR_UNRECOGNIZED_MEDIA

Done:       ret

DMK_Load    endp

;---------------------------------------------------------------------
; DMK_Unload                                         Release interface
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_Unload  proc    pDMK:DWORD

            mcall   DMK_VTBL, SaveTrack, pDMK                           ; Flush the cache contents back to disk

            or      eax, eax
            jnz     Done
            
@@:         mov     edx, pDMK

            fn      GlobalFree, (CDMK PTR [edx]).pTrack                 ; Release allocated memory

            or      eax, eax
            jz      Done

            fn      GetLastError
    
Done:       ret
    
DMK_Unload  endp

;---------------------------------------------------------------------
; DMK_Read                                                 Read sector
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; pBuffer:  Pointer to data buffer
; Len:      Buffer length
; Track:    Track number
; Side:     Side number
; Sector:   Sector number
;---------------------------------------------------------------------
; EAX:      Error code
; ECX:      Sector size
; EDX:      Sector density (0:Single, 1:Double)
;---------------------------------------------------------------------

DMK_Read    proc    pDMK:DWORD, pBuffer:DWORD, Len:DWORD, Track:BYTE, Side:BYTE, Sector:BYTE

            local   SectorHeader:DMK_SectorHeader

    ; --- Validate the arguments ---

            mov     edx, pDMK

            mov     al, Track
            cmp     al, (CDMK PTR [edx]).LastTrack                      ; Track must be less or equal LastTrack
            ja      Error

            mov     al, Side
            cmp     al, (CDMK PTR [edx]).LastSide                       ; Side must be less or equal LastSides
            ja      Error
            
            mov     al, Sector
            cmp     al, (CDMK PTR [edx]).LastSector                     ; Sector must be equal or lower than LastSector
            ja      Error

    ; --- Check whether the track is already loaded ---

            mcall   DMK_VTBL, LoadTrack, pDMK, Track, Side

            or      eax, eax
            jnz     Done
                                                                               
    ; --- Retrieve sector header and data ---

Seek:       mcall   DMK_VTBL, GetSectorID, pDMK, addr SectorHeader, Track, Side, Sector

            or      eax, eax
            jnz     Done

            mcall   DMK_VTBL, GetSectorData, pDMK, addr SectorHeader, pBuffer, Len

            or      eax, eax
            jnz     Done

    ; --- Set the return registers ---

            movzx   ecx, SectorHeader.SectorSize                        ; Get sector size
            
            movzx   edx, SectorHeader.IDAM_PTR                          ; Get sector density
            shr     edx, 15

            mov     eax, NO_ERROR
            jmp     Done

Error:      mov     eax, ERROR_SECTOR_NOT_FOUND

Done:       ret

DMK_Read    endp

;---------------------------------------------------------------------
; DMK_Write                                               Write sector
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; pBuffer:  Pointer to data buffer
; Len:      Buffer length
; Track:    Track number
; Side:     Side number
; Sector:   Sector number
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_Write   proc    pDMK:DWORD, pBuffer:DWORD, Len:DWORD, Track:BYTE, Side:BYTE, Sector:BYTE

            local   SectorHeader:DMK_SectorHeader

    ; --- Validate the arguments ---

            mov     edx, pDMK

            mov     al, Track
            cmp     al, (CDMK PTR [edx]).LastTrack                      ; Track must be less or equal LastTrack
            ja      Error

            mov     al, Side
            cmp     al, (CDMK PTR [edx]).LastSide                       ; Side must be less or equal LastSides
            ja      Error
            
            mov     al, Sector
            cmp     al, (CDMK PTR [edx]).LastSector                     ; Sector must be equal or lower than LastSector
            ja      Error

    ; --- Check whether the track is already loaded ---

            mcall   DMK_VTBL, LoadTrack, pDMK, Track, Side

            or      eax, eax
            jnz     Done
            
    ; --- Update sector ---

            mcall   DMK_VTBL, GetSectorID, pDMK, addr SectorHeader, Track, Side, Sector

            or      eax, eax
            jnz     Done

            mcall   DMK_VTBL, PutSectorData, pDMK, addr SectorHeader, pBuffer, Len

            or      eax, eax
            jnz     Done

            mcall   DMK_VTBL, UpdateCRC, pDMK, addr SectorHeader

    ; --- Set the cache write flag ---
    
            mov     edx, pDMK
            mov     (CDMK PTR [edx]).CacheWrite, 1

            mov     eax, NO_ERROR
            jmp     Done

Error:      mov     eax, ERROR_SECTOR_NOT_FOUND

Done:       ret

DMK_Write   endp

;---------------------------------------------------------------------
; DMK_GetDiskGeometry                       Find out the disk geometry
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
;---------------------------------------------------------------------
; EAX:      Error code (member variables updated directly)
;---------------------------------------------------------------------

DMK_GetDiskGeometry proc    pDMK:DWORD

            local   hFile:HANDLE
            local   pTrack:DWORD
            local   Tracks:BYTE
            local   TrackLength:DWORD
            local   Sides:BYTE
            local   Flags:BYTE
            local   Bytes:DWORD
            local   LowestTrack:BYTE
            local   LowestSide:BYTE
            local   HighestSide:BYTE
            local   LowestSector:BYTE
            local   HighestSector:BYTE
            local   SPT1:BYTE
            local   SPT2:BYTE
            local   Density:BYTE
            local   ScanFlag:BYTE

    ; --- Initialize local variables ---

            mov     LowestTrack, -1
            mov     LowestSide, -1
            mov     HighestSide, 0
            mov     LowestSector, -1
            mov     HighestSector, 0
            mov     SPT1, 0
            mov     SPT2, 0
            mov     Density, -1
            mov     ScanFlag, 0

            mov     edx, pDMK

            mov     eax, (CDMK PTR [edx]).hFile
            mov     hFile, eax

            mov     eax, (CDMK PTR [edx]).pTrack
            mov     pTrack, eax

            mov     al, (CDMK PTR [edx]).Header.Tracks
            mov     Tracks, al

            movzx   eax, (CDMK PTR [edx]).Header.TrackLength
            mov     TrackLength, eax

            test    (CDMK PTR [edx]).Header.Flags, DMK_Flag_SingleSided
            setz    al
            inc     al
            mov     Sides, al
            
            mov     al, (CDMK PTR [edx]).Header.Flags
            mov     Flags, al

    ; --- Load first disk track ---

            mov     eax, sizeof DMK_Header
            
Read:       fn      SetFilePointer, hFile, eax, NULL, FILE_BEGIN

            cmp     eax, -1
            jne     @F

            fn      GetLastError
            jmp     Done
            
@@:         fn      ReadFile, hFile, pTrack, TrackLength, addr Bytes, NULL

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done
            
    ; --- Scan all sector headers for the lowest and highest values ---

@@:         mov     ecx, 64                                             ; Scan up to 64 pointers
            mov     edx, pTrack

Scan:       movzx   eax, word ptr [edx]                                 ; Get IDAM offset

            or      eax, eax                                            ; Check the end of the pointers list
            jz      LastTrack

            push    ecx
            push    edx

    ; --- Determine disk density ---
    
            test    eax, DMK_IDAM_Density                               ; Get sector density
            setnz   dl

            cmp     Density, -1                                         ; Check whether the disk density is init'd
            jne     @F

            mov     Density, dl                                         ; Assume current sector's density
            
@@:         cmp     Density, dl                                         ; Is this a mixed density disk?
            je      @F

            mov     Density, 2                                          ; Set disk density as mixed

    ; --- Get current sector's C/H/S ---

@@:         and     eax, DMK_IDAM_Offset                                ; Convert IDAM offset to a pointer
            add     eax, pTrack

            test    Flags, DMK_Flag_SingleDensity+DMK_Flag_IgnoreDensity    ; Check whether the sector bytes are doubled
            jnz     @F

            or      dl, dl
            jnz     @F

            mov     cl, (DMK_SectorHeader PTR [eax + 1]).SectorID.Track ; Double bytes
            mov     dh, (DMK_SectorHeader PTR [eax + 2]).SectorID.Side
            mov     dl, (DMK_SectorHeader PTR [eax + 3]).SectorID.Sector
            jmp     Compare

@@:         mov     cl, (DMK_SectorHeader PTR [eax]).SectorID.Track     ; Single bytes
            mov     dh, (DMK_SectorHeader PTR [eax]).SectorID.Side
            mov     dl, (DMK_SectorHeader PTR [eax]).SectorID.Sector

    ; --- Compare with current values ---

Compare:    cmp     cl, LowestTrack                                     ; Compare track
            jae     @F

            mov     LowestTrack, cl

@@:         cmp     dh, LowestSide                                      ; Compare side
            jae     @F

            mov     LowestSide, dh

@@:         cmp     dh, HighestSide
            jbe     @F

            mov     HighestSide, dh

@@:         cmp     dl, LowestSector                                    ; Compare sector
            jae     @F

            mov     LowestSector, dl

@@:         cmp     dl, HighestSector
            jbe     @F

            mov     HighestSector, dl

@@:         cmp     ScanFlag, 0                                         ; Count track sectors
            jne     LastTrk

            inc     SPT1                                                ; First track sectors
            jmp     @F
            
LastTrk:    inc     SPT2                                                ; Last track sectors

    ; --- Advance to the next sector ---
    
@@:         pop     edx
            pop     ecx

            add     edx, 2
            dec     ecx
            jnz     Scan

    ; --- Continue comparing at the last disk track ---

LastTrack:  cmp     ScanFlag, 0                                         ; Check whether we have already gone through this
            jne     Update

            inc     ScanFlag

            movzx   eax, Tracks                                         ; Last Track = Tracks * Sides * TrackLength + (sizeof DMK_Head)
            dec     eax
            mul     Sides
            cmp     Sides, 2                                            ; If double sided, advance to the last side
            jb      @F
            inc     eax
@@:         mul     TrackLength
            add     eax, sizeof DMK_Header

            jmp     Read                                                ; Go back and repeat the process

    ; --- Update member variables ---

Update:     mov     edx, pDMK

            mov     al, Sides                                           ; Update number of internal (allocated) sides
            mov     (CDMK PTR [edx]).iSides, al

            mov     al, Tracks                                          ; Update total number of tracks on disk
            mov     (CDMK PTR [edx]).Tracks, al

            or      al, al                                              ; There should be at least one track
            jz      Error

            dec     al
            add     al, LowestTrack                                     ; Update number of last track on disk
            mov     (CDMK PTR [edx]).LastTrack, al
            
            mov     al, LowestTrack                                     ; Update number of first track on disk
            mov     (CDMK PTR [edx]).FirstTrack, al

            mov     al, HighestSide                                     ; Update number of formatted (detected) sides
            sub     al, LowestSide
            inc     al
            mov     (CDMK PTR [edx]).Sides, al

            or      al, al                                              ; There should be at least one side
            jz      Error

            mov     al, LowestSide                                      ; Update number of first disk side
            mov     (CDMK PTR [edx]).FirstSide, al

            mov     al, HighestSide                                     ; Update number of last disk side
            mov     (CDMK PTR [edx]).LastSide, al

            mov     al, SPT1                                            ; Compare first and last track sectors count
            cmp     al, SPT2
            je      @F
            ja      HighestSPT                                          ; If different, keep the highest number
            
            mov     al, SPT2

HighestSPT: mov     Density, 2                                          ; Assure that density is set as mixed

@@:         mov     (CDMK PTR [edx]).SPT, al                            ; Update number of sectors per track

            or      al, al                                              ; There should be at least one sector per track
            jz      Error

            mov     al, LowestSector                                    ; Update number of first sector on track
            mov     (CDMK PTR [edx]).FirstSector, al

            mov     al, HighestSector                                   ; Update number of last sector on track
            mov     (CDMK PTR [edx]).LastSector, al

            mov     al, Density                                         ; Set the disk density
            mov     (CDMK PTR [edx]).Density, al

            mov     eax, NO_ERROR
            jmp     Done

Error:      mov     eax, ERROR_DISK_CORRUPT                             ; The disk structure is corrupted and unreadable.

Done:       ret

DMK_GetDiskGeometry endp

;---------------------------------------------------------------------
; DMK_Seek                                        Compute track offset
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; Track:    Track number
; Side:     Side number
;---------------------------------------------------------------------
; EAX:      Track offset in the image file
;---------------------------------------------------------------------

DMK_Seek    proc    pDMK:DWORD, Track:BYTE, Side:BYTE

            movzx   eax, Track                                          ; Track Offset = ((Track * Sides + Side) * TrackLength) + sizeof DMK_Header

            mov     edx, pDMK
            mul     (CDMK PTR [edx]).iSides

            movzx   ecx, Side
            sub     cl, (CDMK PTR [edx]).FirstSide
            add     eax, ecx

            mul     (CDMK PTR [edx]).Header.TrackLength
            shl     edx, 16
            add     eax, edx

            add     eax, sizeof DMK_Header

            ret
            
DMK_Seek    endp

;---------------------------------------------------------------------
; DMK_LoadTrack                           Load a track into the buffer
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; Track:    Track number
; Side:     Side number
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_LoadTrack   proc    pDMK:DWORD, Track:BYTE, Side:BYTE

            local   Bytes:DWORD

    ; --- Check whether the requested track is already loaded ---

            mov     ah, Track
            mov     al, Side

            cmp     ax, (CDMK PTR [edx]).LoadedTrack
            je      Ready

    ; --- Flush the cache before reloading it ---

            mcall   DMK_VTBL, SaveTrack, pDMK

            or      eax, eax
            jnz     Done
            
    ; --- Compute track offset ---

            mcall   DMK_VTBL, Seek, pDMK, Track, Side                   ; Return track offset in EAX
            
    ; --- Set file pointer ---

            mov     edx, pDMK

            fn      SetFilePointer, (CDMK PTR [edx]).hFile, eax, NULL, FILE_BEGIN

            cmp     eax, -1
            jne     @F
                                                                                                                                
            fn      GetLastError
            jmp     Done

    ; --- Load the track in the buffer ---

@@:         mov     edx, pDMK
            movzx   ecx, (CDMK PTR [edx]).Header.TrackLength

            fn      ReadFile, (CDMK PTR [edx]).hFile, (CDMK PTR [edx]).pTrack, ecx, addr Bytes, NULL

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

    ; --- Update the track caching control ---

@@:         mov     ah, Track
            mov     al, Side

            mov     edx, pDMK
            mov     (CDMK PTR [edx]).LoadedTrack, ax                    ; Update loaded track's Track and Side values

Ready:      mov     eax, NO_ERROR
            jmp     Done

Error:      mov     eax, ERROR_SECTOR_NOT_FOUND

Done:       ret

DMK_LoadTrack   endp

;---------------------------------------------------------------------
; DMK_SaveTrack                  Write the cache contents back to disk
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; Track:    Track number
; Side:     Side number
; Sector:   Sector number
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_SaveTrack   proc    pDMK:DWORD

            local   Bytes:DWORD

    ; --- Check whether the cache write flag is set ---

            mov     edx, pDMK

            cmp     (CDMK PTR [edx]).CacheWrite, 0
            je      Clear

    ; --- Compute track offset ---

            mov     ax, (CDMK PTR [edx]).LoadedTrack                    ; Retrieve currently loaded track values

            movzx   ecx, ah                                             ; Track
            movzx   edx, al                                             ; Side

            mcall   DMK_VTBL, Seek, pDMK, ecx, edx                      ; Return track offset in EAX
            
    ; --- Set file pointer ---

            mov     edx, pDMK

            fn      SetFilePointer, (CDMK PTR [edx]).hFile, eax, NULL, FILE_BEGIN

            cmp     eax, -1
            jne     @F
                                                                                                                                
            fn      GetLastError
            jmp     Done

    ; --- Write track back to disk ---

@@:         mov     edx, pDMK
            movzx   ecx, (CDMK PTR [edx]).Header.TrackLength

            fn      WriteFile, (CDMK PTR [edx]).hFile, (CDMK PTR [edx]).pTrack, ecx, addr Bytes, NULL

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

    ; --- Clear the cache write flag ---

@@:         mov     edx, pDMK
            mov     (CDMK PTR [edx]).CacheWrite, 0

Clear:      mov     eax, NO_ERROR
            
Done:       ret
            
DMK_SaveTrack   endp

;---------------------------------------------------------------------
; DMK_SeekIDAM                                  Locate a sector header
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; Track:    Requested track
; Side:     Requested side
; Sector:   Requested sector
;---------------------------------------------------------------------
; EAX:      Pointer to the sector header or NULL on error
; ECX:      Original track pointer to the sector IDAM
; EDX:      Double byte flag
;---------------------------------------------------------------------

DMK_SeekIDAM        proc    pDMK:DWORD, Track:BYTE, Side:BYTE, Sector:BYTE

            local   IDAM_PTR:WORD
            local   DoubleByte:BYTE

    ; --- Check all IDAM pointers for a matching C/H/S ---

            mov     edx, pDMK
            mov     edx, (CDMK PTR [edx]).pTrack

            xor     ecx, ecx

Seek:       movzx   eax, word ptr [edx + ecx]

            or      eax, eax                                            ; An empty slot indicates the end of the pointers list
            jz      Done

            mov     IDAM_PTR, ax

            and     eax, DMK_IDAM_Offset                                ; Make a pointer to the sector header
            add     eax, edx

            push    ecx
            push    edx

    ; --- Check whether the sector data is doubled ---

            mov     DoubleByte, 0                                       ; Reset the double byte flag

            mov     edx, pDMK

            test    (CDMK PTR [edx]).Header.Flags, DMK_Flag_SingleDensity+DMK_Flag_IgnoreDensity   ; Double density disk?
            jnz     @F

            test    IDAM_PTR, DMK_IDAM_Density                          ; But single density sector?
            jnz     @F

            mov     DoubleByte, 1                                       ; Then set the double byte flag

    ; --- Get current sector C/H/S ---

            mov     cl, (DMK_SectorHeader PTR [eax + 1]).SectorID.Track ; Double bytes
            mov     dh, (DMK_SectorHeader PTR [eax + 2]).SectorID.Side
            mov     dl, (DMK_SectorHeader PTR [eax + 3]).SectorID.Sector
            jmp     Compare

@@:         mov     cl, (DMK_SectorHeader PTR [eax]).SectorID.Track     ; Single bytes
            mov     dh, (DMK_SectorHeader PTR [eax]).SectorID.Side
            mov     dl, (DMK_SectorHeader PTR [eax]).SectorID.Sector

    ; --- Compare current sector with the requested one ---

Compare:
;...........................................................................................
; In some disks the track numbers don't match the real position, so let's ignore this check.
;...........................................................................................
;            cmp     cl, Track                                           ; Same track?
;            jne     Next

;...........................................................................................
; In some disks, sides 0 and 1 are inverted, so checking may prevent reading them.
;...........................................................................................
;            cmp     dh, Side                                            ; Same side?
;            jne     Next

            cmp     dl, Sector                                          ; Same sector?
            je      Found

    ; --- Advance to the next IDAM pointer ---

Next:       pop     edx
            pop     ecx

            add     ecx, 2
            cmp     ecx, 128                                            ; 64 slots max
            jb      Seek

            xor     eax, eax                                            ; Not found, return NULL
            jmp     Done

    ; --- Set return registers ---

Found:      add     esp, 8                                              ; Balance the stack (ECX/EDX)

            movzx   ecx, IDAM_PTR                                       ; Original track pointer to the sector IDAM
            movzx   edx, DoubleByte                                     ; Double byte flag
            
Done:       ret

DMK_SeekIDAM        endp

;---------------------------------------------------------------------
; DMK_GetSectorID                      Locate and load a sector header
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; pDMKHDR:  Pointer to a target sector header
; Track:    Track number
; Side:     Side number
; Sector:   Sector number
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_GetSectorID proc    pDMK:DWORD, pDMKHDR:DWORD, Track:BYTE, Side:BYTE, Sector:BYTE

            local   pIDAM:DWORD
            local   IDAM_PTR:WORD
            local   DoubleByte:BYTE

    ; --- Locate the ID Address Mark ---

            fn      DMK_SeekIDAM, pDMK, Track, Side, Sector

            or      eax, eax                                            ; Return pointer cannot be NULL
            jnz     @F

            mov     eax, ERROR_SECTOR_NOT_FOUND
            jmp     Done

@@:         mov     pIDAM, eax                                          ; EAX = Pointer to the actual sector IDAM
            mov     IDAM_PTR, cx                                        ; ECX = Original track pointer to the sector IDAM
            mov     DoubleByte, dl                                      ; EDX = Double byte flag
                        
    ; --- Copy the ID field to the caller's buffer ---

            push    esi
            push    edi

            mov     esi, eax                                            ; EAX = Pointer to sector header from DMK_SeekIDAM
            mov     edi, pDMKHDR
            lea     edi, (DMK_SectorHeader PTR [edi]).SectorID
            mov     ecx, sizeof DMK_SectorID

@@:         movsb
            add     esi, edx                                            ; Adjust step if double bytes flag is set
            loop    @B

            pop     edi
            pop     esi

    ; --- Set other header variables ---

            mov     edx, pDMKHDR

            mov     ax, IDAM_PTR                                            ; Preserve the original track pointer to the sector IDAM
            mov     (DMK_SectorHeader PTR [edx]).IDAM_PTR, ax

            mov     cl, (DMK_SectorHeader PTR [edx]).SectorID.SectorSize    ; Convert sector size to the actual number of bytes
            mov     ax, 128
            shl     ax, cl
            mov     (DMK_SectorHeader PTR [edx]).SectorSize, ax

            mov     al, DoubleByte                                          ; Save the double byte flag
            mov     (DMK_SectorHeader PTR [edx]).DoubleByte, al

            mov     eax, pIDAM                                              ; Save the actual pointer to the sector IDAM
            mov     (DMK_SectorHeader PTR [edx]).pIDAM, eax

            mov     eax, NO_ERROR

Done:       ret

DMK_GetSectorID endp

;---------------------------------------------------------------------
; DMK_PutSectorID                    Locate and update a sector header
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; pDMKHDR:  Pointer to a source sector header
; Track:    Track number
; Side:     Side number
; Sector:   Sector number
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_PutSectorID proc    pDMK:DWORD, pDMKHDR:DWORD, Track:BYTE, Side:BYTE, Sector:BYTE

    ; --- Locate the ID Address Mark ---

            fn      DMK_SeekIDAM, pDMK, Track, Side, Sector

            or      eax, eax                                            ; Return pointer cannot be NULL
            jnz     @F

            mov     eax, ERROR_SECTOR_NOT_FOUND
            jmp     Done

    ; --- Copy the ID field to the cache ---

@@:         push    esi
            push    edi

            mov     esi, pDMKHDR
            lea     esi, (DMK_SectorHeader PTR [esi]).SectorID
            mov     edi, eax                                            ; EAX = Pointer to sector header from DMK_SeekIDAM
            mov     ecx, sizeof DMK_SectorID

Copy:       lodsb
            stosb

            or      dl, dl                                              ; Copy each byte twice if set
            jz      @F
            
            stosb

@@:         loop    Copy

            pop     edi
            pop     esi

            mov     eax, NO_ERROR

Done:       ret

DMK_PutSectorID endp

;---------------------------------------------------------------------
; DMK_SeekDAM                             Locate the Data Address Mark
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; pDMKHDR:  Pointer to a pre-loaded and target sector header
;---------------------------------------------------------------------
; EAX:      Pointer to the sector DAM or NULL on error
; ECX:      Sector size
; EDX:      Double byte flag
;---------------------------------------------------------------------

DMK_SeekDAM proc    pDMK:DWORD, pDMKHDR:DWORD

    ; --- Compute pointer to sector data ---

            mov     edx, pDMKHDR
            mov     eax, (DMK_SectorHeader PTR [edx]).pIDAM             ; Get the actual pointer to the sector IDAM
            
            mov     cl, (DMK_SectorHeader PTR [edx]).DoubleByte         ; Skip the ID field
            mov     edx, sizeof DMK_SectorID
            shl     edx, cl
            add     eax, edx

    ; --- Look for the Data Address Mark (DAM) ---

            movzx   edx, cl                                             ; Adjust step if double bytes flag is set
            inc     edx
            mov     ecx, 43                                             ; DAM must be found up to 43 bytes after the ID Field

@@:         cmp     byte ptr [eax], 0F8h                                ; Look for Data Address Mark (F8h-FBh)
            jb      Next

            cmp     byte ptr [eax], 0FBh
            jna     Found

Next:       add     eax, edx
            loop    @B

            xor     eax, eax                                            ; Not found, return NULL
            jmp     Done

    ; --- Update the in-memory sector header ---

Found:      mov     edx, pDMKHDR
            mov     (DMK_SectorHeader PTR [edx]).pDAM, eax              ; Save the actual pointer to the sector DAM
            movzx   ecx, (DMK_SectorHeader PTR [edx]).SectorSize        ; Return sector size in ECX
            movzx   edx, (DMK_SectorHeader PTR [edx]).DoubleByte        ; Return double byte flag in EDX
            
Done:       ret

DMK_SeekDAM endp

;---------------------------------------------------------------------
; DMK_GetSectorData                  Retrieve the contents of a sector
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; pDMKHDR:  Pointer to pre-loaded sector header
; pBuffer:  Pointer to source data buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_GetSectorData   proc    pDMK:DWORD, pDMKHDR:DWORD, pBuffer:DWORD, Len:DWORD

    ; --- Locate the Data Address Mark ---

            mcall   DMK_VTBL, SeekDAM, pDMK, pDMKHDR

            or      eax, eax                                            ; Pointer cannot be NULL
            jnz     @F

            mov     eax, ERROR_FLOPPY_ID_MARK_NOT_FOUND
            jmp     Done

    ; --- Check destination buffer size ---

@@:         cmp     ecx, Len                                            ; ECX = sector size
            jbe     @F

            mov     eax, ERROR_INVALID_USER_BUFFER
            jmp     Done

    ; --- Copy sector data to the caller's buffer ---

@@:         push    esi
            push    edi

            mov     edi, pBuffer
            mov     esi, eax                                            ; EAX = Pointer to the Data Address Mark
            inc     esi                                                 ; Skip the DAM byte

@@:         add     esi, edx
            movsb                                                       ; Adjust step if double bytes flag is set
            loop    @B

            pop     edi
            pop     esi

            mov     eax, NO_ERROR

Done:       ret
            
DMK_GetSectorData   endp

;---------------------------------------------------------------------
; DMK_PutSectorData                    Update the contents of a sector
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; pDMKHDR:  Pointer to pre-loaded sector header
; pBuffer:  Pointer to destination data buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

DMK_PutSectorData   proc    pDMK:DWORD, pDMKHDR:DWORD, pBuffer:DWORD, Len:DWORD

    ; --- Locate the Data Address Mark ---

            mcall   DMK_VTBL, SeekDAM, pDMK, pDMKHDR

            or      eax, eax                                            ; Pointer cannot be NULL
            jnz     @F

            mov     eax, ERROR_FLOPPY_ID_MARK_NOT_FOUND
            jmp     Done

    ; --- Check source buffer size ---

@@:         cmp     ecx, Len                                            ; ECX = sector size
            jae     @F

            mov     eax, ERROR_INVALID_USER_BUFFER
            jmp     Done

    ; --- Copy sector data to the cache ---

@@:         push    esi
            push    edi

            mov     esi, pBuffer
            mov     edi, eax                                            ; EAX = Pointer to the Data Address Mark
            inc     edi                                                 ; Skip the DAM byte
            
Copy:       lodsb
            stosb

            or      dl, dl                                              ; Copy each byte twice if double byte flag is set
            jz      @F
            
            stosb

@@:         loop    Copy

            pop     edi
            pop     esi

    ; --- Compute the new CRC ---

            mov     eax, NO_ERROR

Done:       ret

DMK_PutSectorData   endp

;---------------------------------------------------------------------
; DMK_UpdateCRC                                Cyclic Redundancy Check
;---------------------------------------------------------------------
; pDMK:     Pointer to the calling DMK object
; pDMKHDR:  Pointer to pre-loaded sector header
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------
; Algorithm (based on libdmk by Eric Smith):
;   CRC = FFFF
;   For each data byte:
;      data = data SHL 8
;      Repeat 8 times:
;          CRC = (CRC SHL 1) XOR (((CRC XOR data) & 8000) ? 1021 : 0)
;          data = data SHL 1
;---------------------------------------------------------------------

DMK_UpdateCRC   proc    pDMK:DWORD, pDMKHDR:DWORD

            local   String[80]:BYTE
            
            push    ebx
            push    esi
            push    edi

            mov     edx, pDMKHDR
            mov     esi, (DMK_SectorHeader PTR [edx]).pDAM              ; Get actual pointer to DAM byte
            movzx   ecx, (DMK_SectorHeader PTR [edx]).SectorSize        ; Get sector size
            inc     ecx                                                 ; Include DAM in total sector size
            movzx   edi, (DMK_SectorHeader PTR [edx]).DoubleByte        ; Step adjustment

            test    (DMK_SectorHeader PTR [edx]).IDAM_PTR, DMK_IDAM_Density
            jnz     @F
                        
            mov     dx, 0FFFFh                                          ; Initial CRC value for FM (single density)
            jmp     NextByte

@@:         mov     dx, 0CDB4h                                          ; Initial CRC value for MFM (double density) -- CRC of A1 A1 A1

NextByte:   push    ecx
            lodsb
            shl     ax, 8                                               ; data = data SHL 8
            mov     cx, 8                                               ; Repeat 8 times:
NextBit:    mov     bx, dx
            xor     bx, ax                                              ; (CRC XOR data)
            shl     dx, 1                                               ; CRC = CRC SHL 1
            bt      bx, 15                                              ; MSB set?
            jnc     @F                                                  ; XOR against polynom
            xor     dx, 1021h
@@:         shl     ax, 1                                               ; data = data SHL 1
            loop    NextBit                                             ; Next bit
            add     esi, edi
            pop     ecx
            loop    NextByte                                            ; Next byte

            or      di, di                                              ; Check double byte flag
            jz      @F

            mov     [esi+0], dh                                         ; Update CRC (double byte)
            mov     [esi+1], dh
            mov     [esi+2], dl
            mov     [esi+3], dl
            jmp     Updated
            
@@:         mov     [esi+0], dh                                         ; Update CRC (single byte)
            mov     [esi+1], dl

Updated:    pop     edi
            pop     esi
            pop     ebx

Done:       ret

DMK_UpdateCRC   endp

