;---------------------------------------------------------------------
; JV1 Virtual Disk Interface                Copyright (C) Miguel Dutra
;---------------------------------------------------------------------

.data?

;---------------------------------------------------------------------
; JV1 Class
;---------------------------------------------------------------------

CJV1    struct
                        CVDI        <>                                  ; Inherit all VDI class members
    Sectors             WORD        ?                                   ; Total number of sectors on disk
CJV1    ends

.data

;---------------------------------------------------------------------
; JV1 Virtual Table
;---------------------------------------------------------------------

JV1_VTBL    struct                                                      ; Declare the structure and initializers
    Load                DWORD       JV1_Load                            ; VDI-derived functions
    Unload              DWORD       JV1_Unload
    Read                DWORD       JV1_Read
    Write               DWORD       JV1_Write
    GetTracks           DWORD       VDI_GetTracks
    GetSides            DWORD       VDI_GetSides
    GetSPT              DWORD       VDI_GetSPT
    GetDensity          DWORD       VDI_GetDensity
    GetFirstTrack       DWORD       VDI_GetFirstTrack
    GetLastTrack        DWORD       VDI_GetLastTrack
    GetFirstSide        DWORD       VDI_GetFirstSide
    GetLastSide         DWORD       VDI_GetLastSide
    GetFirstSector      DWORD       VDI_GetFirstSector
    GetLastSector       DWORD       VDI_GetLastSector
    Seek                DWORD       JV1_Seek                            ; JV1-specific function
JV1_VTBL    ends

JV1                     JV1_VTBL    <>                                  ; Define a variable of that type

.code
    
;---------------------------------------------------------------------
; JV1_Load                                        Initialize interface
;---------------------------------------------------------------------
; pJV1:     Pointer to the calling JV1 object
; hFile:    Handle to an open Virtual Floppy Disk file
; Flags:    User-defined behavior modification flags
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

JV1_Load    proc    pJV1:DWORD, hFile:HANDLE, Flags:DWORD

            local   Sectors:WORD
            local   Tracks:BYTE
            local   Sides:BYTE
            local   SectorsPerTrack:BYTE
            
    ; --- Find out the total number of disk sectors ---

            fn      GetFileSize, hFile, NULL                            ; Get the file size

            or      al, al                                              ; Must be multiple of 256 bytes (last byte of size = 0)
            jnz     IntError

            shr     eax, 8                                              ; Divide it by 256 to get the number of disk sectors
            mov     Sectors, ax

    ; --- Determine the disk geometry ---
    
            cmp     ax, 400                                             ; 100KB disk (40 tracks, 10 sectors per track, single sided)
            ja      @F

            mov     SectorsPerTrack, 10
            div     SectorsPerTrack
            mov     Tracks, al
            mov     Sides, 1
            jmp     Check

@@:         cmp     ax, 720                                             ; 180KB disk (40 tracks, 18 sectors per track, single sided)
            ja      @F

            mov     SectorsPerTrack, 18
            div     SectorsPerTrack
            mov     Tracks, al
            mov     Sides, 1
            jmp     Check

@@:         mov     SectorsPerTrack, 18                                 ; All others (40+ tracks, 18 sectors per track, double sided)
            div     SectorsPerTrack
            shr     al, 1
            mov     Tracks, al
            mov     Sides, 2

Check:      or      ah, ah                                              ; There must be no remainder after last division
            jnz     IntError

    ; --- Update member variables ---
    
            mov     edx, pJV1    

            mov     eax, hFile                                          ; Save the file handle
            mov     (CJV1 PTR [edx]).hFile, eax

            mov     eax, Flags                                          ; Update the behavior modification flags
            mov     (CJV1 PTR [edx]).Flags, eax

            mov     al, Tracks                                          ; Update number of Tracks and Last Track
            mov     (CJV1 PTR [edx]).Tracks, al
            dec     al
            mov     (CJV1 PTR [edx]).LastTrack, al

            mov     al, Sides                                           ; Update number of Sides and Last Side
            mov     (CJV1 PTR [edx]).Sides, al
            dec     al
            mov     (CJV1 PTR [edx]).LastSide, al

            mov     al, SectorsPerTrack                                 ; Update number of Sectors Per Track and Last Sector
            mov     (CJV1 PTR [edx]).SPT, al
            dec     al
            mov     (CJV1 PTR [edx]).LastSector, al

            mov     (CJV1 PTR [edx]).Density, 0                         ; Update Density, First Track, First Side and First Sector
            mov     (CJV1 PTR [edx]).FirstTrack, 0
            mov     (CJV1 PTR [edx]).FirstSide, 0
            mov     (CJV1 PTR [edx]).FirstSector, 0

            mov     ax, Sectors                                         ; Update the total number of disk sectors
            mov     (CJV1 PTR [edx]).Sectors, ax

            mov     eax, NO_ERROR
            jmp     Done
            
IntError:   mov     eax, ERROR_UNRECOGNIZED_MEDIA

Done:       ret

JV1_Load    endp

;---------------------------------------------------------------------
; JV1_Unload                                         Release interface
;---------------------------------------------------------------------
; pJV1: Pointer to the calling JV1 object
;---------------------------------------------------------------------
; EAX:  Error code
;---------------------------------------------------------------------

JV1_Unload  proc    pJV1:DWORD

            mov     eax, NO_ERROR
            ret

JV1_Unload  endp

;---------------------------------------------------------------------
; JV1_Read                                                 Read sector
;---------------------------------------------------------------------
; pJV1:     Pointer to the calling JV1 object
; pBuffer:  Pointer to data buffer
; Len:      Buffer length
; Track:    Track number
; Side:     Side number
; Sector:   Sector number
;---------------------------------------------------------------------
; EAX:      Error code
; ECX:      Sector size
; EDX:      Sector density (0:Single, 1:Double)
;---------------------------------------------------------------------

JV1_Read    proc    pJV1:DWORD, pBuffer:DWORD, Len:DWORD, Track:BYTE, Side:BYTE, Sector:BYTE

            local   Bytes:DWORD
            
    ; --- Set file pointer ---

            mcall   JV1_VTBL, Seek, pJV1, Track, Side, Sector

            or      eax, eax
            jnz     Done
            
    ; --- Check buffer size ---

            cmp     Len, 256
            jbe     @F

            mov     Len, 256

    ; --- Read sector from disk ---

@@:         mov     edx, pJV1

            fn      ReadFile, (CJV1 PTR [edx]).hFile, pBuffer, Len, addr Bytes, NULL

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

    ; --- Set return registers ---
    
@@:         mov     eax, NO_ERROR
            mov     ecx, 256                                            ; Set sector size
            mov     edx, 0                                              ; Set sector density (always single?)

Done:       ret

JV1_Read    endp

;---------------------------------------------------------------------
; JV1_Write                                               Write sector
;---------------------------------------------------------------------
; pJV1:     Pointer to the calling JV1 object
; pBuffer:  Pointer to data buffer
; Len:      Buffer length
; Track: 	Track number
; Side:     Side number
; Sector:   Sector number
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

JV1_Write   proc    pJV1:DWORD, pBuffer:DWORD, Len:DWORD, Track:BYTE, Side:BYTE, Sector:BYTE

            local   Bytes:DWORD
            
    ; --- Set file pointer ---

            mcall   JV1_VTBL, Seek, pJV1, Track, Side, Sector

            or      eax, eax
            jnz     Done
            
    ; --- Check data length ---

            cmp     Len, 256
            jbe     @F

            mov     Len, 256

    ; --- Write sector to disk ---

@@:         mov     edx, pJV1

            fn      WriteFile, (CJV1 PTR [edx]).hFile, pBuffer, Len, addr Bytes, NULL

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     eax, NO_ERROR

Done:       ret

JV1_Write   endp

;---------------------------------------------------------------------
; JV1_Seek                      Set file pointer to a requested sector
;---------------------------------------------------------------------
; pJV1:     Pointer to the calling JV1 object
; Track:    Track number
; Side:     Side number
; Sector:   Sector number
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

JV1_Seek    proc    pJV1:DWORD, Track:BYTE, Side:BYTE, Sector:BYTE

    ; --- Validate the arguments ---

            mov     edx, pJV1

            mov     al, Track                                           ; Validate Track
            cmp     al, (CJV1 PTR [edx]).Tracks
            jnb     Error

            mov     al, Side                                            ; Validate Side
            cmp     al, (CJV1 PTR [edx]).Sides
            jnb     Error

            mov     al, Sector                                          ; Validate Sector
            cmp     al, (CJV1 PTR [edx]).SPT
            jnb     Error

    ; --- Compute sector offset ---

            movzx   eax, (CJV1 PTR [edx]).Sides                         ; Offset = ((Track * Sides + Side) * SectorsPerTrack + Sector) * 256
            mul     Track
            add     al, Side
            mul     (CJV1 PTR [edx]).SPT
            movzx   ecx, Sector
            add     eax, ecx
            shl     eax, 8
            
    ; --- Set file pointer ---

            mov     edx, pJV1

            fn      SetFilePointer, (CJV1 PTR [edx]).hFile, eax, NULL, FILE_BEGIN

            cmp     eax, -1
            jne     @F

            fn      GetLastError
            jmp     Done

@@:         mov     eax, NO_ERROR
            jmp     Done

Error:      mov     eax, ERROR_SECTOR_NOT_FOUND

Done:       ret

JV1_Seek    endp

